# -*- coding: utf-8 -*-
"""
Created on Thu Nov 22 23:45:09 2018

@author: stasb
"""

import pkg_resources
import numpy as np
import pandas as pd
import math
import warnings
from numba import compiler, types
from scipy.optimize import root
from .integrators import base_integrator, dop853_integrator
from .solout import default_solout

class base_model:
    '''
    Class base_model is intended to provide interface to all presented or
    future models used in orbipy.
    Model consists of but not limited to: 
        - physical constants,
        - reference to integrator object,
        - ...
    '''
    def __init__(self):
        self._constants = np.zeros((1,))
        self.integrator = None
        self.columns = []
        self.stm = False
        
    @property
    def constants(self):
        return self._constants
    
    @constants.setter
    def constants(self, constants):
        self._constants = constants
        
    def get_state_size(self, stm=None):
        pass
    
    def get_zero_state(self):
        pass
    
    def prop(self, s0, t0, t1, ret_df=True):
        pass
    
    def to_df(self):
        return pd.DataFrame()
        
    def __repr__(self):
        return "base_model class instance"

class nondimensional_model(base_model):
    '''
    Class nondimensional_model is intended to provide interface to all
    presented or future nondimensional models used in orbipy.
    '''
    def get_nd_coefs(self):
        return {'T':1., 'L':1.}

def load_constants_csv(fname):
    with open(fname, 'rt') as f:
        df = pd.read_csv(f, index_col=0)
    return df

class crtbp3_model(nondimensional_model):
    '''
    Class crtbp3_model presents Circular Restricted Three-Body Problem model
    in nondimensional formulation (see Murray, Dermott "Solar System Dynamics").
    
    Example
    -------
    
    
    '''
    constants_csv = pkg_resources.resource_filename(__name__, 'data/crtbp_constants.csv')
    constants_df = load_constants_csv(constants_csv)
    
    columns = ['t', 'x', 'y', 'z', 'vx', 'vy', 'vz']
    columns_stm = columns + ["stm%d%d"%(i,j) for i in range(6) for j in range(6)]
    
    ode_compiled = None
    ode_stm_compiled = None

    def __init__(self, const_set='Sun-Earth (default)', 
                       integrator=None, 
                       solout=default_solout(), stm=False):
        '''
        Initializes crtbp3_model instance using name of a constant set,
        reference to integrator and reference to solout callback.
        
        Parameters
        ----------
        
        const_set : str
            Constant set name. Following will be loaded from
            'data/crtbp_constants.csv' file:
                - mu - gravitational constant, mu = m/(M+m), \
                    where m is smaller mass, M is bigger mass;
                - R - distance between primaries in km;
                - T - synodic period of primaries in seconds.
                
        integrator : object
            Object used for numerical intergration of model equation set.
            Should inherit from base_integrator.
            By default: dop853_integrator
            
        solout : object
            Solout callback object will be used for gathering all 
            integration steps from scipy.integrate.ode integrators
            and for event detection routines.
            Should inherit from base_solout.
            By default: default_solout which can only gather all
            integration steps.
        '''
#        doesn't work with russian symbols in file path        
#        self.df = pd.read_csv(self.constants_csv, index_col=0)
        self._constants = np.array(self.constants_df.loc[const_set,'mu':'T'].values, dtype=float)
        self.M = self.constants_df.loc[const_set, 'M']
        self.m = self.constants_df.loc[const_set, 'm']
        self.mu = self._constants[0]
        self.mu1 = 1.0 - self.mu
        self.R = self._constants[1]
        self.T = self._constants[2]
        self.const_set = const_set
        if integrator is None:
            self.integrator = dop853_integrator()
        else:
            if not isinstance(integrator, base_integrator):
                raise TypeError('integrator should be an instance ob base_integrator')
            self.integrator = integrator
        self.solout = solout
        self.stm = stm
        self.columns = crtbp3_model.columns_stm if self.stm else crtbp3_model.columns
        if stm:
            if crtbp3_model.ode_stm_compiled is None:
                crtbp3_model.ode_stm_compiled = \
                compiler.compile_isolated(crtbp3_model.crtbp_ode_stm, 
                                          [types.double, types.double[:], types.double[:]], 
                                          return_type=types.double[:]).entry_point
            self.right_part = crtbp3_model.ode_stm_compiled
        else:
            if crtbp3_model.ode_compiled is None:
                crtbp3_model.ode_compiled = \
                compiler.compile_isolated(crtbp3_model.crtbp_ode, 
                                          [types.double, types.double[:], types.double[:]], 
                                          return_type=types.double[:]).entry_point
            self.right_part = crtbp3_model.ode_compiled
            
        self.L = self.lagrange_points()
        self.L1 = self.L[0, 0]
        self.L2 = self.L[1, 0]
        self.L3 = self.L[2, 0]

    def get_nd_coefs(self):
        '''
        Calculates nondimensional coefficients which is used by scaler
        for conversion between nondimensional units and physical.
        '''
        return {'T':self.T / (2*np.pi),
                'L':self.R}
        
    @property
    def solout(self):
        '''
        solout property getter
        '''
        return self._solout
    
    @solout.setter
    def solout(self, solout):
        '''
        solout property setter
        '''
        if not solout:
            raise RuntimeWarning('Incorrect solout function/object')
            self._solout = default_solout()
        else:
            self._solout = solout    
        if self.integrator:
            self.integrator.solout = self._solout
        

    def lagrange_points(self):
        '''
        Numerically calculate position of all 5 Lagrange points.
        
        Returns
        -------
        
        L : np.array of (5, 3) shape
            Positions of Lagrange points for this model.
        '''
        def opt(x, constants):
            s = self.get_zero_state()
            s[0] = x[0]
            return self.right_part(0., s, constants)[3]
        
        L = np.zeros((5, 3))
        L[0, 0] = root(opt,  0.5, args=(self._constants,)).x[0]
        L[1, 0] = root(opt,  2.0, args=(self._constants,)).x[0]
        L[2, 0] = root(opt, -1.0, args=(self._constants,)).x[0]
        L[3, 0] = 0.5 - self.mu
        L[3, 1] = 0.5*3**0.5
        L[4, 0] = 0.5 - self.mu
        L[4, 1] = -0.5*3**0.5
        return L

    def omega(self, s):
        if s.ndim == 1:
            r1 = ((s[0] + self.mu )**2 + s[1]**2 + s[2]**2)**0.5
            r2 = ((s[0] - self.mu1)**2 + s[1]**2 + s[2]**2)**0.5
            return 0.5*(s[0]**2 + s[1]**2) + self.mu1/r1 + self.mu/r2        
        else:
            r1 = np.sqrt((s[:,0] + self.mu )**2 + s[:,1]**2 + s[:,2]**2)
            r2 = np.sqrt((s[:,0] - self.mu1)**2 + s[:,1]**2 + s[:,2]**2)
            return 0.5*(s[:,0]**2 + s[:,1]**2) + self.mu1/r1 + self.mu/r2
        
    def jacobi(self, s):
        if s.ndim == 1:
            return 2*self.omega(s) - s[3]**2 - s[4]**2 - s[5]**2
        else:
            return 2*self.omega(s) - s[:,3]**2 - s[:,4]**2 - s[:,5]**2

    #@staticmethod
    def crtbp_ode(t, s, constants):
        '''
        Right part of CRTPB ODE
    
        Parameters
        ----------
        t : scalar
            Nondimensional time (same as angle of system rotation).
    
        s : np.array with 6 components
            State vector of massless spacecraft (x,y,z,vx,vy,vz).
    
        constants : np.array
             mu = constants[0] - gravitaional parameter of crtbp model
    
        Returns
        -------
    
        ds : np.array
            First order derivative with respect to time of spacecraft
            state vector (vx,vy,vz,dvx,dvy,dvz)
        '''
        mu2 = constants[0]
        mu1 = 1 - mu2
        
        x, y, z, vx, vy, vz = s
        
        yz2 = y * y + z * z
        r13 = ((x + mu2) * (x + mu2) + yz2) ** (-1.5)
        r23 = ((x - mu1) * (x - mu1) + yz2) ** (-1.5)
    
        mu12r12 = (mu1 * r13 + mu2 * r23)
    
        ax =  2 * vy + x - (mu1 * (x + mu2) * r13 + mu2 * (x - mu1) * r23)
        ay = -2 * vx + y - mu12r12 * y
        az =             - mu12r12 * z
        
        ds = np.array([vx, vy, vz, ax, ay, az])
        
        return ds

#    crtbp_ode_compiled = \
#    compiler.compile_isolated(crtbp_ode, 
#                              [types.double, types.double[:], types.double[:]], 
#                              return_type=types.double[:]).entry_point

    #@staticmethod    
    def crtbp_ode_stm(t, s, constants):
        '''
        Right part of CRTPB ODE with State Transform Matrix calculation.
    
        Parameters
        ----------
        t : scalar
            Nondimensional time (same as angle of system rotation).
    
        s : np.array with 42 (=6+6*6) components
            State vector of massless spacecraft (x,y,z,vx,vy,vz) along with
            flattened STM matrix (stm00,stm01,stm02,...stm55).
    
        constants : np.array
             mu = constants[0] - gravitaional parameter of crtbp model
    
        Returns
        -------
    
        ds : np.array
            First order derivative with respect to time of spacecraft
            state vector along with flattened derivative of STM matrix
            (vx,vy,vz,dvx,dvy,dvz,dstm00,dstm01,...,dstm55)
        '''
    
        x, y, z, vx, vy, vz = s[:6]
        stm0 = np.ascontiguousarray(s[6:]).reshape(6,6)
        mu2 = constants[0]
        mu1 = 1 - mu2
        
        yz2 = y * y + z * z;
        xmu2 = x + mu2
        xmu1 = x - mu1
    
        r1 = (xmu2**2 + yz2)**0.5
        r2 = (xmu1**2 + yz2)**0.5
        r13, r15 = r1**(-3), r1**(-5)
        r23, r25 = r2**(-3), r2**(-5)
    
        mu12r12 = (mu1 * r13 + mu2 * r23);
    
        ax =  2. * vy + x - (mu1 * xmu2 * r13 + mu2 * xmu1 * r23);
        ay = -2. * vx + y - mu12r12 * y;
        az =              - mu12r12 * z;
            
        Uxx = 1. - mu12r12 + 3*mu1*xmu2**2*r15 + 3*mu2*xmu1**2*r25
        Uxy = 3*mu1*xmu2*y*r15 + 3*mu2*xmu1*y*r25
        Uxz = 3*mu1*xmu2*z*r15 + 3*mu2*xmu1*z*r25
        Uyy = 1. - mu12r12 + 3*mu1*y**2*r15 + 3*mu2*y**2*r25
        Uyz = 3*mu1*y*z*r15 + 3*mu2*y*z*r25
        Uzz = -mu12r12 + 3*mu1*z**2*r15 + 3*mu2*z**2*r25
        
        A = np.array(((0.,  0.,  0.,   1., 0., 0.),
                      (0.,  0.,  0.,   0., 1., 0.),
                      (0.,  0.,  0.,   0., 0., 1.),
                      (Uxx, Uxy, Uxz,  0., 2., 0.),
                      (Uxy, Uyy, Uyz, -2., 0., 0.),
                      (Uxz, Uyz, Uzz,  0., 0., 0.)))
    
        stm1 = np.dot(A,stm0)
        ds = np.empty_like(s)
        ds[0], ds[1], ds[2], ds[3], ds[4], ds[5] = vx, vy, vz, ax, ay, az
        ds[6:] = stm1.ravel()
        
        return ds

#    crtbp_ode_stm_compiled = \
#    compiler.compile_isolated(crtbp_ode_stm, 
#                              [types.double, types.double[:], types.double[:]], 
#                              return_type=types.double[:]).entry_point

    def get_state_size(self, stm=None):
        '''
        Calculate size of a spacecraft state vector.
        '''
        if stm is None:
            return 42 if self.stm else 6
        else:
            return 42 if stm else 6
    
    def get_zero_state(self):
        '''
        Returns
        -------
            zs : numpy.ndarray
                State vector of appropriate size filled with zeros.
                If state contains STM then STM is filled by eye matrix.
        '''
        if self.stm:
            zs = np.zeros(42)
            zs[6::7] = 1.0 # set eye matrix
            return zs
        else:
            return np.zeros(6)
    
    def prop(self, s0, t0, t1, ret_df=True):
        '''
        Propagate spacecraft from initial state s0 at time t0 up to time t1.
        
        Parameters
        ----------
        
        s0 : np.array
            Spacecraft initial state of size that match model
            
        t0 : float
            Initial time
            
        t1 : float
            Boundary time
            
        ret_df : bool
            If True, returns pd.DataFrame
            Else, returns np.array
            
        Returns
        -------
            df : np.array (ret_df=False) of pd.DataFrame (ret_df=True)
                Array or DataFrame consists of spacecraft states for all
                steps made by integrator from t0 to t1.
        
        '''
        self._solout.reset()
        self.integrator.integrate(self.right_part, s0, t0, t1, fargs=(self._constants,))
        if ret_df:
            return self.to_df(self._solout.out)
        else:
            return np.array(self._solout.out)
    
    def to_df(self, arr, columns=None):#, index_col='t'):
        if columns is None:
            columns = self.columns
        df = pd.DataFrame(arr, columns=columns)
#        df.set_index(index_col, inplace=True)
        return df
    
    def __repr__(self):
        return 'CRTBP3 model%s:%r'%(' with STM' if self.stm else '', 
                                    self.const_set)
    

class crtbp3_custom_model(nondimensional_model):
    '''
    Class crtbp3_custom_model presents Circular Restricted Three-Body Problem model
    in nondimensional formulation (see Murray, Dermott "Solar System Dynamics").
    This model have customizable right-part-ode function and should be built by 
    crtbp3_model_builder.
    
    Example
    -------
    
    See model_builder.py
    
    '''
    constants_csv = pkg_resources.resource_filename(__name__, 'data/crtbp_constants.csv')
    constants_df = load_constants_csv(constants_csv)
    
    def __init__(self, right_part, #state_size=6, stm_size=36, 
                       ev_size=0,
                       const_set=None, 
                       integrator=None, 
                       solout=None,
                       stm=False):
        '''
        Initializes crtbp3_model instance using name of a constant set,
        reference to integrator and reference to solout callback.
        
        Parameters
        ----------
        
        const_set : str
            Constant set name. Following will be loaded from
            'data/crtbp_constants.csv' file:
                - mu - gravitational constant, mu = m/(M+m), \
                    where m is smaller mass, M is bigger mass;
                - R - distance between primaries in km;
                - T - synodic period of primaries in seconds.
                
        integrator : object
            Object used for numerical intergration of model equation set.
            Should inherit from base_integrator.
            By default: dop853_integrator
            
        solout : object
            Solout callback object will be used for gathering all 
            integration steps from scipy.integrate.ode integrators
            and for event detection routines.
            Should inherit from base_solout.
            By default: default_solout which can only gather all
            integration steps.
        '''
#        doesn't work with russian symbols in file path        
#        self.df = pd.read_csv(self.constants_csv, index_col=0)
        if const_set is None:
            const_set = self.constants_df.index[0]
        self._constants = np.array(self.constants_df.loc[const_set,'mu':'T'].values, dtype=float)
        self.M = self.constants_df.loc[const_set, 'M']
        self.m = self.constants_df.loc[const_set, 'm']
        self.mu = self._constants[0]
        self.mu1 = 1.0 - self.mu
        self.R = self._constants[1]
        self.T = self._constants[2]
        self.const_set = const_set
        if integrator is None:
            self.integrator = dop853_integrator()
        else:
            if not isinstance(integrator, base_integrator):
                raise TypeError('integrator should be an instance ob base_integrator')
            self.integrator = integrator
        self.solout = solout
        self.stm = stm
        self.ev_size = ev_size
        
        columns = ['t', 'x', 'y', 'z', 'vx', 'vy', 'vz']
        columns_stm = ["stm%d%d"%(i,j) for i in range(6) for j in range(6)]
        columns_ev = ["ev%d"%i for i in range(self.ev_size)]
        
        self.columns = columns + (columns_stm if self.stm else []) + columns_ev
        self.right_part = right_part            
        self.L = self.lagrange_points()
        self.L1 = self.L[0, 0]
        self.L2 = self.L[1, 0]
        self.L3 = self.L[2, 0]

    def get_nd_coefs(self):
        '''
        Calculates nondimensional coefficients which is used by scaler
        for conversion between nondimensional units and physical.
        '''
        return {'T':self.T / (2*np.pi),
                'L':self.R}
        
    @property
    def solout(self):
        '''
        solout property getter
        '''
        return self._solout
    
    @solout.setter
    def solout(self, solout):
        '''
        solout property setter
        '''
        if not solout:
#            warnings.warn('Incorrect solout function/object')
            self._solout = default_solout()
        else:
            self._solout = solout    
        if self.integrator:
            self.integrator.solout = self._solout
        

    def lagrange_points(self):
        '''
        Numerically calculate position of all 5 Lagrange points.
        
        Returns
        -------
        
        L : np.array of (5, 3) shape
            Positions of Lagrange points for this model.
        '''
        def opt(x, constants):
            s = self.get_zero_state()
            s[0] = x[0]
            return self.right_part(0., s, constants)[3]
        
        L = np.zeros((5, 3))
        L[0, 0] = root(opt,  0.5, args=(self._constants,)).x[0]
        L[1, 0] = root(opt,  2.0, args=(self._constants,)).x[0]
        L[2, 0] = root(opt, -1.0, args=(self._constants,)).x[0]
        L[3, 0] = 0.5 - self.mu
        L[3, 1] = 0.5*3**0.5
        L[4, 0] = 0.5 - self.mu
        L[4, 1] = -0.5*3**0.5
        return L

    def omega(self, s):
        if s.ndim == 1:
            r1 = ((s[0] + self.mu )**2 + s[1]**2 + s[2]**2)**0.5
            r2 = ((s[0] - self.mu1)**2 + s[1]**2 + s[2]**2)**0.5
            return 0.5*(s[0]**2 + s[1]**2) + self.mu1/r1 + self.mu/r2        
        else:
            r1 = np.sqrt((s[:,0] + self.mu )**2 + s[:,1]**2 + s[:,2]**2)
            r2 = np.sqrt((s[:,0] - self.mu1)**2 + s[:,1]**2 + s[:,2]**2)
            return 0.5*(s[:,0]**2 + s[:,1]**2) + self.mu1/r1 + self.mu/r2
        
    def jacobi(self, s):
        if s.ndim == 1:
            return 2*self.omega(s) - s[3]**2 - s[4]**2 - s[5]**2
        else:
            return 2*self.omega(s) - s[:,3]**2 - s[:,4]**2 - s[:,5]**2

    def get_state_size(self, detailed=False):
        '''
        Calculate size of a spacecraft state vector.
        '''
        if detailed:
            return (6, (36 if self.stm else 0), self.ev_size)
        else:
            return 6 + (36 if self.stm else 0) + self.ev_size
    
    def get_zero_state(self):
        '''
        Returns
        -------
            zs : numpy.ndarray
                State vector of appropriate size filled with zeros.
                If state contains STM then STM is filled by eye matrix.
        '''      
        fs = self.get_state_size()
        zs = np.zeros(fs)
        
        if self.stm:
            zs[6:42:7] = 1.0 # set eye matrix
            
        return zs
    
    def prop(self, s0, t0, t1, ret_df=True):
        '''
        Propagate spacecraft from initial state s0 at time t0 up to time t1.
        
        Parameters
        ----------
        
        s0 : np.array
            Spacecraft initial state of size that match model
            
        t0 : float
            Initial time
            
        t1 : float
            Boundary time
            
        ret_df : bool
            If True, returns pd.DataFrame
            Else, returns np.array
            
        Returns
        -------
            df : np.array (ret_df=False) of pd.DataFrame (ret_df=True)
                Array or DataFrame consists of spacecraft states for all
                steps made by integrator from t0 to t1.
        
        '''
        self._solout.reset()
        self.integrator.integrate(self.right_part, s0, t0, t1, fargs=(self._constants,))
        if ret_df:
            return self.to_df(self._solout.out)
        else:
            return np.array(self._solout.out)
    
    def to_df(self, arr, columns=None):#, index_col='t'):
        if columns is None:
            columns = self.columns
        df = pd.DataFrame(arr, columns=columns)
#        df.set_index(index_col, inplace=True)
        return df
    
    def __repr__(self):
        return 'Custom CRTBP3 model%s:%r, state size: %d,'%\
                (' with STM' if self.stm else '', 
                 self.const_set,
                 self.get_state_size())

    
#class _crtbp3_model_experimental(nondimensional_model):
#    '''
#    Class crtbp3_model presents Circular Restricted Three-Body Problem model
#    in nondimensional formulation (see Murray, Dermott "Solar System Dynamics").
#    
#    Example
#    -------
#    
#    
#    '''
#    constants_csv = pkg_resources.resource_filename(__name__, 'data/crtbp_constants.csv')
#    constants_df = load_constants_csv(constants_csv)
#    
#    columns = ['t', 'x', 'y', 'z', 'vx', 'vy', 'vz']
#    columns_stm = columns + ["stm%d%d"%(i,j) for i in range(6) for j in range(6)]
#    
#    ode_compiled = None
#    ode_stm_compiled = None
#
#    def __init__(self, const_set='Sun-Earth (default)', 
#                       integrator=None, 
#                       solout=default_solout(), stm=False):
#        '''
#        Initializes crtbp3_model instance using name of a constant set,
#        reference to integrator and reference to solout callback.
#        
#        Parameters
#        ----------
#        
#        const_set : str
#            Constant set name. Following will be loaded from
#            'data/crtbp_constants.csv' file:
#                - mu - gravitational constant, mu = m/(M+m), \
#                    where m is smaller mass, M is bigger mass;
#                - R - distance between primaries in km;
#                - T - synodic period of primaries in seconds.
#                
#        integrator : object
#            Object used for numerical intergration of model equation set.
#            Should inherit from base_integrator.
#            By default: dop853_integrator
#            
#        solout : object
#            Solout callback object will be used for gathering all 
#            integration steps from scipy.integrate.ode integrators
#            and for event detection routines.
#            Should inherit from base_solout.
#            By default: default_solout which can only gather all
#            integration steps.
#        '''
##        doesn't work with russian symbols in file path        
##        self.df = pd.read_csv(self.constants_csv, index_col=0)
#        self._constants = np.array(self.constants_df.loc[const_set,'mu':'T'].values, dtype=float)
#        self.M = self.constants_df.loc[const_set, 'M']
#        self.m = self.constants_df.loc[const_set, 'm']
#        self.mu = self._constants[0]
#        self.mu1 = 1.0 - self.mu
#        self.R = self._constants[1]
#        self.T = self._constants[2]
#        self.const_set = const_set
#        if integrator is None:
#            self.integrator = dop853_integrator()
#        else:
#            if not isinstance(integrator, base_integrator):
#                raise TypeError('integrator should be an instance ob base_integrator')
#            self.integrator = integrator
#        self.solout = solout
#        self.stm = stm
#        self.columns = crtbp3_model.columns_stm if self.stm else crtbp3_model.columns
#        if stm:
#            if _crtbp3_model_experimental.ode_stm_compiled is None:
#                _crtbp3_model_experimental.ode_stm_compiled = \
#                compiler.compile_isolated(_crtbp3_model_experimental.crtbp_ode_stm, 
#                                          [types.double, types.double[:], types.double[:]], 
#                                          return_type=types.double[:]).entry_point
#            self.right_part = _crtbp3_model_experimental.ode_stm_compiled
#        else:
#            if _crtbp3_model_experimental.ode_compiled is None:
#                _crtbp3_model_experimental.ode_compiled = \
#                compiler.compile_isolated(_crtbp3_model_experimental.crtbp_ode, 
#                                          [types.double, types.double[:], types.double[:]], 
#                                          return_type=types.double[:]).entry_point
#            self.right_part = _crtbp3_model_experimental.ode_compiled
#            
#        self.L = self.lagrange_points()
#        self.L1 = self.L[0, 0]
#        self.L2 = self.L[1, 0]
#        self.L3 = self.L[2, 0]
#
#    def get_nd_coefs(self):
#        '''
#        Calculates nondimensional coefficients which is used by scaler
#        for conversion between nondimensional units and physical.
#        '''
#        return {'T':self.T / (2*np.pi),
#                'L':self.R}
#        
#    @property
#    def solout(self):
#        '''
#        solout property getter
#        '''
#        return self._solout
#    
#    @solout.setter
#    def solout(self, solout):
#        '''
#        solout property setter
#        '''
#        if not solout:
#            raise RuntimeWarning('Incorrect solout function/object')
#            self._solout = default_solout()
#        else:
#            self._solout = solout    
#        if self.integrator:
#            self.integrator.solout = self._solout
#        
#
#    def lagrange_points(self):
#        '''
#        Numerically calculate position of all 5 Lagrange points.
#        
#        Returns
#        -------
#        
#        L : np.array of (5, 3) shape
#            Positions of Lagrange points for this model.
#        '''
#        def opt(x, constants):
#            s = self.get_zero_state()
#            s[0] = x[0]
#            return self.right_part(0., s, constants)[3]
#        
#        L = np.zeros((5, 3))
#        L[0, 0] = root(opt,  0.5, args=(self._constants,)).x[0]
#        L[1, 0] = root(opt,  2.0, args=(self._constants,)).x[0]
#        L[2, 0] = root(opt, -1.0, args=(self._constants,)).x[0]
#        L[3, 0] = 0.5 - self.mu
#        L[3, 1] = 0.5*3**0.5
#        L[4, 0] = 0.5 - self.mu
#        L[4, 1] = -0.5*3**0.5
#        return L
#
#    def omega(self, s):
#        if s.ndim == 1:
#            r1 = ((s[0] + self.mu )**2 + s[1]**2 + s[2]**2)**0.5
#            r2 = ((s[0] - self.mu1)**2 + s[1]**2 + s[2]**2)**0.5
#            return 0.5*(s[0]**2 + s[1]**2) + self.mu1/r1 + self.mu/r2        
#        else:
#            r1 = np.sqrt((s[:,0] + self.mu )**2 + s[:,1]**2 + s[:,2]**2)
#            r2 = np.sqrt((s[:,0] - self.mu1)**2 + s[:,1]**2 + s[:,2]**2)
#            return 0.5*(s[:,0]**2 + s[:,1]**2) + self.mu1/r1 + self.mu/r2
#        
#    def jacobi(self, s):
#        if s.ndim == 1:
#            return 2*self.omega(s) - s[3]**2 - s[4]**2 - s[5]**2
#        else:
#            return 2*self.omega(s) - s[:,3]**2 - s[:,4]**2 - s[:,5]**2
#
##    #@staticmethod
#    def crtbp_ode(t, s, constants):
#        '''
#        Right part of CRTPB ODE
#    
#        Parameters
#        ----------
#        t : scalar
#            Nondimensional time (same as angle of system rotation).
#    
#        s : np.array with 6 components
#            State vector of massless spacecraft (x,y,z,vx,vy,vz).
#    
#        constants : np.array
#             mu = constants[0] - gravitaional parameter of crtbp model
#    
#        Returns
#        -------
#    
#        ds : np.array
#            First order derivative with respect to time of spacecraft
#            state vector (vx,vy,vz,dvx,dvy,dvz)
#        '''
#        mu2 = constants[0]
#        mu1 = 1 - mu2
#        
#        x, y, z, vx, vy, vz, _, _ = s
#        
#        yz2 = y * y + z * z
#        r13 = ((x + mu2) * (x + mu2) + yz2) ** (-1.5)
#        r23 = ((x - mu1) * (x - mu1) + yz2) ** (-1.5)
#    
#        mu12r12 = (mu1 * r13 + mu2 * r23)
#    
#        ax =  2 * vy + x - (mu1 * (x + mu2) * r13 + mu2 * (x - mu1) * r23)
#        ay = -2 * vx + y - mu12r12 * y
#        az =             - mu12r12 * z
#        
##        def impulse(x, kx, ky):
##            x1 = -x/kx
##            if x1 > 100.0: 
##                x1 = 100.0
##            elif x1 < -100.0:
##                x1 = -100.0
##            e = math.exp(x1)
##            return -ky*e*(e-1.0)/(1.0 + e)**3
#
#        def dgauss(x, a, b, c, d):
#            xb = x-b
#            c2 = c**2
#            return -a*xb*math.exp(-math.fabs(xb)**d/(2*c2))/c2
#        
#        # event perception part
#        
#        d = 1.5
#        
#        ky0 = 1e-6
#        kx0 = 1e-3
#        x01 = 1.009029061194469
#        x02 = 1.0110344087880518
##        evx = impulse(x-x01, kx0, ky0) + impulse(x-x02, kx0, ky0)
#        evx1 = dgauss(x, ky0, x01, kx0, d)
#        evx2 = dgauss(x, ky0, x02, kx0, d)
#
#        ky1 = 1e-7
#        kx1 = 1e-3
##        evy = impulse(y, kx1, ky1)
#        evy = dgauss(y, ky1, 0.0, kx1, d)
#        
##        evx = 0
##        evy = 0
#        ds = np.array([vx, vy, vz, ax, ay, az, evx1, evx2])
#        
#        return ds
##
###    crtbp_ode_compiled = crtbp_ode
##
##    crtbp_ode_compiled = \
##    compiler.compile_isolated(crtbp_ode, 
##                              [types.double, types.double[:], types.double[:]], 
##                              return_type=types.double[:]).entry_point
##
##    #@staticmethod    
#    def crtbp_ode_stm(t, s, constants):
#        '''
#        Right part of CRTPB ODE with State Transform Matrix calculation.
#    
#        Parameters
#        ----------
#        t : scalar
#            Nondimensional time (same as angle of system rotation).
#    
#        s : np.array with 42 (=6+6*6) components
#            State vector of massless spacecraft (x,y,z,vx,vy,vz) along with
#            flattened STM matrix (stm00,stm01,stm02,...stm55).
#    
#        constants : np.array
#             mu = constants[0] - gravitaional parameter of crtbp model
#    
#        Returns
#        -------
#    
#        ds : np.array
#            First order derivative with respect to time of spacecraft
#            state vector along with flattened derivative of STM matrix
#            (vx,vy,vz,dvx,dvy,dvz,dstm00,dstm01,...,dstm55)
#        '''
#    
#        x, y, z, vx, vy, vz = s[:6]
#        stm0 = np.ascontiguousarray(s[6:]).reshape(6,6)
#        mu2 = constants[0]
#        mu1 = 1 - mu2
#        
#        yz2 = y * y + z * z;
#        xmu2 = x + mu2
#        xmu1 = x - mu1
#    
#        r1 = (xmu2**2 + yz2)**0.5
#        r2 = (xmu1**2 + yz2)**0.5
#        r13, r15 = r1**(-3), r1**(-5)
#        r23, r25 = r2**(-3), r2**(-5)
#    
#        mu12r12 = (mu1 * r13 + mu2 * r23);
#    
#        ax =  2. * vy + x - (mu1 * xmu2 * r13 + mu2 * xmu1 * r23);
#        ay = -2. * vx + y - mu12r12 * y;
#        az =              - mu12r12 * z;
#            
#        Uxx = 1. - mu12r12 + 3*mu1*xmu2**2*r15 + 3*mu2*xmu1**2*r25
#        Uxy = 3*mu1*xmu2*y*r15 + 3*mu2*xmu1*y*r25
#        Uxz = 3*mu1*xmu2*z*r15 + 3*mu2*xmu1*z*r25
#        Uyy = 1. - mu12r12 + 3*mu1*y**2*r15 + 3*mu2*y**2*r25
#        Uyz = 3*mu1*y*z*r15 + 3*mu2*y*z*r25
#        Uzz = -mu12r12 + 3*mu1*z**2*r15 + 3*mu2*z**2*r25
#        
#        A = np.array(((0.,  0.,  0.,   1., 0., 0.),
#                      (0.,  0.,  0.,   0., 1., 0.),
#                      (0.,  0.,  0.,   0., 0., 1.),
#                      (Uxx, Uxy, Uxz,  0., 2., 0.),
#                      (Uxy, Uyy, Uyz, -2., 0., 0.),
#                      (Uxz, Uyz, Uzz,  0., 0., 0.)))
#    
#        stm1 = np.dot(A,stm0)
#        ds = np.empty_like(s)
#        ds[0], ds[1], ds[2], ds[3], ds[4], ds[5] = vx, vy, vz, ax, ay, az
#        ds[6:] = stm1.ravel()
#        
#        return ds
##
##    crtbp_ode_stm_compiled = \
##    compiler.compile_isolated(crtbp_ode_stm, 
##                              [types.double, types.double[:], types.double[:]], 
##                              return_type=types.double[:]).entry_point
#
#    def get_state_size(self, stm=None):
#        '''
#        Calculate size of a spacecraft state vector.
#        '''
#        if stm is None:
#            return 44 if self.stm else 8
#        else:
#            return 44 if stm else 8
#    
#    def get_zero_state(self):
#        '''
#        Returns
#        -------
#            zs : numpy.ndarray
#                State vector of appropriate size filled with zeros.
#                If state contains STM then STM is filled by eye matrix.
#        '''
#        if self.stm:
#            zs = np.zeros(44)
#            zs[8::7] = 1.0 # set eye matrix
#            return zs
#        else:
#            return np.zeros(8)
#    
#    def prop(self, s0, t0, t1, ret_df=True):
#        '''
#        Propagate spacecraft from initial state s0 at time t0 up to time t1.
#        
#        Parameters
#        ----------
#        
#        s0 : np.array
#            Spacecraft initial state of size that match model
#            
#        t0 : float
#            Initial time
#            
#        t1 : float
#            Boundary time
#            
#        ret_df : bool
#            If True, returns pd.DataFrame
#            Else, returns np.array
#            
#        Returns
#        -------
#            df : np.array (ret_df=False) of pd.DataFrame (ret_df=True)
#                Array or DataFrame consists of spacecraft states for all
#                steps made by integrator from t0 to t1.
#        
#        '''
#        self._solout.reset()
#        self.integrator.integrate(self.right_part, s0, t0, t1, fargs=(self._constants,))
#        if ret_df:
#            return self.to_df(self._solout.out)
#        else:
#            return np.array(self._solout.out)
#    
#    def to_df(self, arr, columns=None):#, index_col='t'):
#        if columns is None:
#            columns = self.columns
#        df = pd.DataFrame(arr, columns=columns)
##        df.set_index(index_col, inplace=True)
#        return df
#    
#    def __repr__(self):
#        return 'CRTBP3 model%s:%r'%(' with STM' if self.stm else '', 
#                                    self.const_set)
        


# -*- coding: utf-8 -*-
"""
Created on Tue Nov 27 21:16:30 2018

@author: maria
"""

import numpy as np
import math
from .events import event_detector, eventY
from .mapper import mapper

class differential_correction:
    '''
    differential_correction class itroduce differential
    correction method: 
        ds = Phi(t0,t)*ds0,
        where:
        Phi(t0,t) - State Transition Matrix from t0 to t,
        ds0 - variation of state at t0,
        ds - variation of state at t.
        
    Thus, ds0 = Phi(t0,t)**(-1)*ds.
    This set of equations can be solved partially when some
    assumptions is made.
    
    For example, to find halo orbits one can assume:
        vx(t), vz(t) should be zero (target),
        and x(t0), vy(t0) will vary
    
    Then correction will be calculated as:
        
        m = [[Phi[3,0], Phi[3,4]], [Phi[5,0], Phi[5,4]]]
        -1/vy(t)*dot([[ax(t)],[az(t)]],[Phi[1,0], Phi[1,4]])
        
        dx = inv(m)*(-[vx(t),vz(t)])
        
    '''
    presets = {
                'halo': {'target':('vx','vz'),
                         'vary':('x','z','vy'), 
                         'event': eventY()},
                'axial':{'target':('z','vx'),
                         'vary':('vy','vy','vz'),
                         'event':eventY(count=2)}
                }
    
    def __init__(self, model,
                 target=('vx','vz'),
                 vary=('x','z','vy'),
                 event=None,
                 iterations=30, 
                 tol=1e-12,
                 ftol=1e-12,
                 maxt=100*np.pi):
        self.target = target
        self.vary = vary
        self.rows = np.array([mapper.col2idx(c)-1 for c in target], dtype=int)
        self.cols = np.array([mapper.col2idx(c)-1 for c in vary], dtype=int)
        self.iterations = iterations
        self.tol = tol
        self.ftol = ftol
        self.maxt = maxt
        if event is None:
            self.event = differential_correction.presets['halo']['event']
        else:
            self.event = event
        self.model = model
        self.it = 0
        
    @classmethod
    def from_preset(cls, model,
                    preset='halo',
                    iterations=30, 
                    tol=1e-12,
                    ftol=1e-12,
                    maxt=100*np.pi):
        if preset not in differential_correction.presets:
            raise RuntimeError("Can't find differential correction preset: %s"%preset)
        return cls(model, **differential_correction.presets[preset],
                   iterations=iterations, tol=tol, ftol=ftol, maxt=maxt)

    @property
    def model(self):
        return self._model
    
    @model.setter
    def model(self, model):
        if not model.stm:
            raise ValueError('Differential correction needs state-transition matrix (STM)')
        self._model = model
        self._detector = event_detector(self._model, [self.event], self.tol)
          
    def shoot(self, s0, alpha=0):
        
        rads = math.radians(alpha)
        self.it = 0
        s = self._model.get_zero_state()
        s[:s0.shape[0]] = s0
        #s = s0.copy()
        r = self.rows
        rT = r[:,np.newaxis]
        c = self.cols[np.newaxis,:]
        
        sn = math.sin(rads)
        cs = math.cos(rads)
        rot = np.array([[cs, sn]]).T
        
        while self.it < self.iterations:
            _, evout = self._detector.prop(s, 0., self.maxt, ret_df=False)
            
            if len(evout) == 0:
                raise RuntimeError('Event unreachable: %r'%self.event)
            
            ev = evout[-1, 4:] # i(0), cnt(1), trm(2), t(3), x(4), y(5), z(6), ...
            xf = ev[:6]
            if math.fabs(xf[r[0]]) < self.ftol and math.fabs(xf[r[1]]) < self.ftol:
                break
            dxdt = self._model.right_part(0., ev, self._model.constants)
            Phi = ev[6:42].reshape(6, 6)
            
            Phi1 = Phi[rT, c]
            Phi2 = np.hstack((Phi1[:,[0,1]]@rot, Phi1[:,[2]]))
            Phi3 = np.hstack((Phi[1,c[:,[0,1]]]@rot, Phi[1, c[:,[2]]]))
            m = Phi2 - 1/dxdt[1] * dxdt[rT]@Phi3
            dd = np.linalg.inv(m)@(-xf[rT])
            dx = np.empty((3,1))
            dx[0] = dd[0,0]*cs
            dx[1] = dd[0,0]*sn
            dx[2] = dd[1,0]
            s[c.T] += dx
#            print(dx)

            self.it += 1

        if self.it >= self.iterations:
            raise RuntimeError("Unable to converge")

        return s
        
    # deprecated
    def find_halo(self, s0, fixed='z0', ret_it=False, ftol=1e-12):
        i = 0
        s = s0.copy()
        while i < self.iterations:
            _, evout = self._detector.prop(s, 0., 700*np.pi, ret_df=False)
#            print(evout[-1, 2])
            ev = evout[-1, 4:] # i(0), cnt(1), trm(2), t(3), x(4), y(5), z(6), ...
            xf = ev[:6]
#            print(xf)
            if math.fabs(xf[3]) < ftol and math.fabs(xf[5]) < ftol:
                break
#            print('shape:', ev.shape)
            dxdt = self._model.right_part(0., ev, self._model.constants)
#            print(xf, dxdt)
            Phi = ev[6:42].reshape(6, 6)
        
            if fixed == 'x0':
                # A = np.array([[Phi[3,2], Phi[3,4]], [Phi[5,2], Phi[5,4]]])
                m = np.array([[Phi[3,2], Phi[3,4]], [Phi[5,2], Phi[5,4]]])\
                -1/dxdt[1]*np.array([[dxdt[3]],[dxdt[5]]])@np.array([Phi[1,2], Phi[1,4]], ndmin=2)
                dx = np.linalg.solve(m, -xf[[3,5]])
                s[[2,4]] += dx
                
            if fixed == 'z0':
                m = np.array([[Phi[3,0], Phi[3,4]], [Phi[5,0], Phi[5,4]]])\
                -1/dxdt[1]*np.array([[dxdt[3]],[dxdt[5]]])@np.array([Phi[1,0], Phi[1,4]], ndmin=2)
#                print(m)
                dx = np.linalg.solve(m, -xf[[3,5]])
#                print(dx)
                s[[0,4]] += dx 
    
            i += 1
        if i < self.iterations:
            pass
#            print("Converged i =", i, "halo", s)
        else:
            raise RuntimeError("Unable to converge")
#            print("Hasn't converged")
        if ret_it:
            return s, i
        return s
    
    
    # deprecated
    def find_axial(self, s0, retIt=False, verbose=False, ftol=1e-12):
        detector = event_detector(self._model, [eventY(count=2)], self.tol)
#        dt = 0
        i = 0
        while i < self.iterations:
#            lst_xf = []
            evout = []
            s = s0.copy()
            _, evout = detector.prop(s, 0, 700*np.pi, ret_df=False)
            ev = evout[-1, 4:] # i(0), cnt(1), trm(2), t(3), x(4), y(5), z(6), ...
            xf = ev[:6]
            dxdt = self._model.right_part(0., ev, self._model.constants)
            Phi = ev[6:42].reshape(6, 6)
            
            if abs(xf[2]) < ftol and abs(xf[3]) < ftol:
                break

            m = np.array([[Phi[2,4], Phi[2,5]], [Phi[3,4], Phi[3,5]]])\
            -1/dxdt[1]*np.array([[dxdt[2]],[dxdt[3]]])@np.array([Phi[1,4], Phi[1,5]], ndmin=2)

            dx = np.linalg.inv(m)@(-xf[[2,3]])
            s0[[4,5]] += dx 

            i += 1
        if verbose:
            if i < self.iterations:
                print("Converged i =", i, "axial", s0)    
            else:
                print("Hasn't converged")
        if retIt:
            return s0.copy(), i
        return s0.copy()
    
    


# -*- coding: utf-8 -*-
"""
Created on Sun Dec  2 13:36:50 2018

@author: stasb
"""

__name__ = "orbipy"

from .models import (crtbp3_model, 
                     crtbp3_custom_model)

from .model_builder import crtbp3_model_builder

from .integrators import (dop853_integrator, 
                          dopri5_integrator, 
                          quasi_integrator)

from .events import *
from .plotter import plotter, plottable, point
from .scaler import scaler
from .mapper import mapper

from . import lines

from .stepper import (stepper, 
                      iteration_printer)

from .extrapolator import (const_extrapolator, 
                           linear_extrapolator, 
                           quadratic_extrapolator, 
                           cubic_extrapolator)

from .interpolators import (linear_interpolator,
                            log_interpolator,
                            ABab_mapper,
                            flipX_interpolator,
                            flipY_interpolator,
                            chain_interpolator)

from .directions import (x_direction,
                         y_direction,
                         xy_angle_direction, 
                         velocity_direction, 
                         unstable_direction,
                         unstable_direction_stm)

from .corrections import (border_correction, 
                          max_time_correction)

from .station_keeping import (simple_station_keeping,
                              strict_station_keeping,
                              montecarlo_station_keeping)

from .manifold import manifold

from .diff_corr import differential_correction

from .multiprocessor import mp